# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: process_parasitics.proto
# Protobuf Python Version: 5.29.0
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    5,
    29,
    0,
    '',
    'process_parasitics.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x18process_parasitics.proto\x12\tkpex.tech\"\x8a\x01\n\x15ProcessParasiticsInfo\x12\x11\n\tside_halo\x18\n \x01(\x05\x12-\n\nresistance\x18n \x01(\x0b\x32\x19.kpex.tech.ResistanceInfo\x12/\n\x0b\x63\x61pacitance\x18o \x01(\x0b\x32\x1a.kpex.tech.CapacitanceInfo\"\x98\x02\n\x0eResistanceInfo\x12\x39\n\x06layers\x18\x01 \x03(\x0b\x32).kpex.tech.ResistanceInfo.LayerResistance\x12\x35\n\x04vias\x18\x02 \x03(\x0b\x32\'.kpex.tech.ResistanceInfo.ViaResistance\x1a]\n\x0fLayerResistance\x12\x12\n\nlayer_name\x18\x01 \x01(\t\x12\x12\n\nresistance\x18\x02 \x01(\x01\x12\"\n\x1a\x63orner_adjustment_fraction\x18\x03 \x01(\x01\x1a\x35\n\rViaResistance\x12\x10\n\x08via_name\x18\x01 \x01(\t\x12\x12\n\nresistance\x18\x02 \x01(\x01\"\x98\x05\n\x0f\x43\x61pacitanceInfo\x12\x44\n\nsubstrates\x18\xc8\x01 \x03(\x0b\x32/.kpex.tech.CapacitanceInfo.SubstrateCapacitance\x12@\n\x08overlaps\x18\xc9\x01 \x03(\x0b\x32-.kpex.tech.CapacitanceInfo.OverlapCapacitance\x12\x42\n\tsidewalls\x18\xca\x01 \x03(\x0b\x32..kpex.tech.CapacitanceInfo.SidewallCapacitance\x12H\n\x0csideoverlaps\x18\xcb\x01 \x03(\x0b\x32\x31.kpex.tech.CapacitanceInfo.SideOverlapCapacitance\x1a\x63\n\x14SubstrateCapacitance\x12\x12\n\nlayer_name\x18\x01 \x01(\t\x12\x18\n\x10\x61rea_capacitance\x18\x02 \x01(\x01\x12\x1d\n\x15perimeter_capacitance\x18\x03 \x01(\x01\x1a\\\n\x12OverlapCapacitance\x12\x16\n\x0etop_layer_name\x18\x01 \x01(\t\x12\x19\n\x11\x62ottom_layer_name\x18\x02 \x01(\t\x12\x13\n\x0b\x63\x61pacitance\x18\x03 \x01(\x01\x1aN\n\x13SidewallCapacitance\x12\x12\n\nlayer_name\x18\x01 \x01(\t\x12\x13\n\x0b\x63\x61pacitance\x18\x02 \x01(\x01\x12\x0e\n\x06offset\x18\x03 \x01(\x01\x1a\\\n\x16SideOverlapCapacitance\x12\x15\n\rin_layer_name\x18\x01 \x01(\t\x12\x16\n\x0eout_layer_name\x18\x02 \x01(\t\x12\x13\n\x0b\x63\x61pacitance\x18\x03 \x01(\x01\"\x0e\n\x0cStyleVariantb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'process_parasitics_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  DESCRIPTOR._loaded_options = None
  _globals['_PROCESSPARASITICSINFO']._serialized_start=40
  _globals['_PROCESSPARASITICSINFO']._serialized_end=178
  _globals['_RESISTANCEINFO']._serialized_start=181
  _globals['_RESISTANCEINFO']._serialized_end=461
  _globals['_RESISTANCEINFO_LAYERRESISTANCE']._serialized_start=313
  _globals['_RESISTANCEINFO_LAYERRESISTANCE']._serialized_end=406
  _globals['_RESISTANCEINFO_VIARESISTANCE']._serialized_start=408
  _globals['_RESISTANCEINFO_VIARESISTANCE']._serialized_end=461
  _globals['_CAPACITANCEINFO']._serialized_start=464
  _globals['_CAPACITANCEINFO']._serialized_end=1128
  _globals['_CAPACITANCEINFO_SUBSTRATECAPACITANCE']._serialized_start=761
  _globals['_CAPACITANCEINFO_SUBSTRATECAPACITANCE']._serialized_end=860
  _globals['_CAPACITANCEINFO_OVERLAPCAPACITANCE']._serialized_start=862
  _globals['_CAPACITANCEINFO_OVERLAPCAPACITANCE']._serialized_end=954
  _globals['_CAPACITANCEINFO_SIDEWALLCAPACITANCE']._serialized_start=956
  _globals['_CAPACITANCEINFO_SIDEWALLCAPACITANCE']._serialized_end=1034
  _globals['_CAPACITANCEINFO_SIDEOVERLAPCAPACITANCE']._serialized_start=1036
  _globals['_CAPACITANCEINFO_SIDEOVERLAPCAPACITANCE']._serialized_end=1128
  _globals['_STYLEVARIANT']._serialized_start=1130
  _globals['_STYLEVARIANT']._serialized_end=1144
# @@protoc_insertion_point(module_scope)
