# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: tech.proto
# Protobuf Python Version: 5.29.0
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    5,
    29,
    0,
    '',
    'tech.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


import process_stack_pb2 as process__stack__pb2
import process_parasitics_pb2 as process__parasitics__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\ntech.proto\x12\tkpex.tech\x1a\x13process_stack.proto\x1a\x18process_parasitics.proto\"\xef\x01\n\nTechnology\x12\x0c\n\x04name\x18\x01 \x01(\t\x12$\n\x06layers\x18\x65 \x03(\x0b\x32\x14.kpex.tech.LayerInfo\x12\x39\n\x13lvs_computed_layers\x18x \x03(\x0b\x32\x1c.kpex.tech.ComputedLayerInfo\x12\x33\n\rprocess_stack\x18\x8c\x01 \x01(\x0b\x32\x1b.kpex.tech.ProcessStackInfo\x12=\n\x12process_parasitics\x18\xc8\x01 \x01(\x0b\x32 .kpex.tech.ProcessParasiticsInfo\"W\n\tLayerInfo\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x13\n\x0b\x64\x65scription\x18\x0b \x01(\t\x12\x11\n\tgds_layer\x18\x15 \x01(\r\x12\x14\n\x0cgds_datatype\x18\x1f \x01(\r\"\xf0\x01\n\x11\x43omputedLayerInfo\x12/\n\x04kind\x18\n \x01(\x0e\x32!.kpex.tech.ComputedLayerInfo.Kind\x12(\n\nlayer_info\x18\x14 \x01(\x0b\x32\x14.kpex.tech.LayerInfo\x12\x1b\n\x13original_layer_name\x18\x1e \x01(\t\"c\n\x04Kind\x12\x14\n\x10KIND_UNSPECIFIED\x10\x00\x12\x10\n\x0cKIND_REGULAR\x10\x01\x12\x19\n\x15KIND_DEVICE_CAPACITOR\x10\x02\x12\x18\n\x14KIND_DEVICE_RESISTOR\x10\x03\x62\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'tech_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  DESCRIPTOR._loaded_options = None
  _globals['_TECHNOLOGY']._serialized_start=73
  _globals['_TECHNOLOGY']._serialized_end=312
  _globals['_LAYERINFO']._serialized_start=314
  _globals['_LAYERINFO']._serialized_end=401
  _globals['_COMPUTEDLAYERINFO']._serialized_start=404
  _globals['_COMPUTEDLAYERINFO']._serialized_end=644
  _globals['_COMPUTEDLAYERINFO_KIND']._serialized_start=545
  _globals['_COMPUTEDLAYERINFO_KIND']._serialized_end=644
# @@protoc_insertion_point(module_scope)
