# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: process_stack.proto
# Protobuf Python Version: 5.29.0
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    5,
    29,
    0,
    '',
    'process_stack.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x13process_stack.proto\x12\tkpex.tech\"\xb5\x0f\n\x10ProcessStackInfo\x12\x35\n\x06layers\x18\x64 \x03(\x0b\x32%.kpex.tech.ProcessStackInfo.LayerInfo\x1a?\n\x07\x43ontact\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x13\n\x0bmetal_above\x18\n \x01(\t\x12\x11\n\tthickness\x18\x14 \x01(\x01\x1a\x46\n\x0eSubstrateLayer\x12\x0e\n\x06height\x18\x01 \x01(\x01\x12\x11\n\tthickness\x18\x02 \x01(\x01\x12\x11\n\treference\x18\x1e \x01(\t\x1ak\n\nNWellLayer\x12\x0e\n\x06height\x18\x01 \x01(\x01\x12\x11\n\treference\x18\x1e \x01(\t\x12:\n\rcontact_above\x18( \x01(\x0b\x32#.kpex.tech.ProcessStackInfo.Contact\x1ao\n\x0e\x44iffusionLayer\x12\x0e\n\x06height\x18\x01 \x01(\x01\x12\x11\n\treference\x18\x1e \x01(\t\x12:\n\rcontact_above\x18( \x01(\x0b\x32#.kpex.tech.ProcessStackInfo.Contact\x1a\'\n\x0f\x46ieldOxideLayer\x12\x14\n\x0c\x64ielectric_k\x18\n \x01(\x01\x1a@\n\x15SimpleDielectricLayer\x12\x14\n\x0c\x64ielectric_k\x18\n \x01(\x01\x12\x11\n\treference\x18\x1e \x01(\t\x1a\x9f\x01\n\x18\x43onformalDielectricLayer\x12\x14\n\x0c\x64ielectric_k\x18\n \x01(\x01\x12\x1c\n\x14thickness_over_metal\x18\x14 \x01(\x01\x12 \n\x18thickness_where_no_metal\x18\x15 \x01(\x01\x12\x1a\n\x12thickness_sidewall\x18\x16 \x01(\x01\x12\x11\n\treference\x18\x1e \x01(\t\x1a~\n\x17SidewallDielectricLayer\x12\x14\n\x0c\x64ielectric_k\x18\n \x01(\x01\x12\x1a\n\x12height_above_metal\x18\x14 \x01(\x01\x12\x1e\n\x16width_outside_sidewall\x18\x15 \x01(\x01\x12\x11\n\treference\x18\x1e \x01(\t\x1a\x9d\x01\n\nMetalLayer\x12\x0e\n\x06height\x18\x01 \x01(\x01\x12\x11\n\tthickness\x18\x02 \x01(\x01\x12\x17\n\x0freference_below\x18\x1e \x01(\t\x12\x17\n\x0freference_above\x18\x1f \x01(\t\x12:\n\rcontact_above\x18( \x01(\x0b\x32#.kpex.tech.ProcessStackInfo.Contact\x1a\xc4\x05\n\tLayerInfo\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x39\n\nlayer_type\x18\x02 \x01(\x0e\x32%.kpex.tech.ProcessStackInfo.LayerType\x12\x45\n\x0fsubstrate_layer\x18Z \x01(\x0b\x32*.kpex.tech.ProcessStackInfo.SubstrateLayerH\x00\x12=\n\x0bnwell_layer\x18\t \x01(\x0b\x32&.kpex.tech.ProcessStackInfo.NWellLayerH\x00\x12\x45\n\x0f\x64iffusion_layer\x18\n \x01(\x0b\x32*.kpex.tech.ProcessStackInfo.DiffusionLayerH\x00\x12H\n\x11\x66ield_oxide_layer\x18\x0b \x01(\x0b\x32+.kpex.tech.ProcessStackInfo.FieldOxideLayerH\x00\x12T\n\x17simple_dielectric_layer\x18\x0c \x01(\x0b\x32\x31.kpex.tech.ProcessStackInfo.SimpleDielectricLayerH\x00\x12Z\n\x1a\x63onformal_dielectric_layer\x18\r \x01(\x0b\x32\x34.kpex.tech.ProcessStackInfo.ConformalDielectricLayerH\x00\x12X\n\x19sidewall_dielectric_layer\x18\x0e \x01(\x0b\x32\x33.kpex.tech.ProcessStackInfo.SidewallDielectricLayerH\x00\x12=\n\x0bmetal_layer\x18\x0f \x01(\x0b\x32&.kpex.tech.ProcessStackInfo.MetalLayerH\x00\x42\x0c\n\nparameters\"\x8e\x02\n\tLayerType\x12\x1a\n\x16LAYER_TYPE_UNSPECIFIED\x10\x00\x12\x18\n\x14LAYER_TYPE_SUBSTRATE\x10\n\x12\x14\n\x10LAYER_TYPE_NWELL\x10\x14\x12\x18\n\x14LAYER_TYPE_DIFFUSION\x10\x1e\x12\x1a\n\x16LAYER_TYPE_FIELD_OXIDE\x10(\x12 \n\x1cLAYER_TYPE_SIMPLE_DIELECTRIC\x10\x32\x12#\n\x1fLAYER_TYPE_CONFORMAL_DIELECTRIC\x10<\x12\"\n\x1eLAYER_TYPE_SIDEWALL_DIELECTRIC\x10\x46\x12\x14\n\x10LAYER_TYPE_METAL\x10Pb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'process_stack_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  DESCRIPTOR._loaded_options = None
  _globals['_PROCESSSTACKINFO']._serialized_start=35
  _globals['_PROCESSSTACKINFO']._serialized_end=2008
  _globals['_PROCESSSTACKINFO_CONTACT']._serialized_start=110
  _globals['_PROCESSSTACKINFO_CONTACT']._serialized_end=173
  _globals['_PROCESSSTACKINFO_SUBSTRATELAYER']._serialized_start=175
  _globals['_PROCESSSTACKINFO_SUBSTRATELAYER']._serialized_end=245
  _globals['_PROCESSSTACKINFO_NWELLLAYER']._serialized_start=247
  _globals['_PROCESSSTACKINFO_NWELLLAYER']._serialized_end=354
  _globals['_PROCESSSTACKINFO_DIFFUSIONLAYER']._serialized_start=356
  _globals['_PROCESSSTACKINFO_DIFFUSIONLAYER']._serialized_end=467
  _globals['_PROCESSSTACKINFO_FIELDOXIDELAYER']._serialized_start=469
  _globals['_PROCESSSTACKINFO_FIELDOXIDELAYER']._serialized_end=508
  _globals['_PROCESSSTACKINFO_SIMPLEDIELECTRICLAYER']._serialized_start=510
  _globals['_PROCESSSTACKINFO_SIMPLEDIELECTRICLAYER']._serialized_end=574
  _globals['_PROCESSSTACKINFO_CONFORMALDIELECTRICLAYER']._serialized_start=577
  _globals['_PROCESSSTACKINFO_CONFORMALDIELECTRICLAYER']._serialized_end=736
  _globals['_PROCESSSTACKINFO_SIDEWALLDIELECTRICLAYER']._serialized_start=738
  _globals['_PROCESSSTACKINFO_SIDEWALLDIELECTRICLAYER']._serialized_end=864
  _globals['_PROCESSSTACKINFO_METALLAYER']._serialized_start=867
  _globals['_PROCESSSTACKINFO_METALLAYER']._serialized_end=1024
  _globals['_PROCESSSTACKINFO_LAYERINFO']._serialized_start=1027
  _globals['_PROCESSSTACKINFO_LAYERINFO']._serialized_end=1735
  _globals['_PROCESSSTACKINFO_LAYERTYPE']._serialized_start=1738
  _globals['_PROCESSSTACKINFO_LAYERTYPE']._serialized_end=2008
# @@protoc_insertion_point(module_scope)
